import sqlite3, logging, os


# Configure logging
logger = logging.getLogger(__name__)


class Database:
    def __init__(self, db_name):
        # Initialize the database connection
        logger.info("Initializing database connection...")
        self.connection = sqlite3.connect(db_name, check_same_thread=False)
        self.cursor = self.connection.cursor()
        logger.info("Database connection established.")

        # Create users table if it doesn't exist
        logger.info("Creating users table if it doesn't exist...")
        self.cursor.execute('''
            CREATE TABLE IF NOT EXISTS users (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                username TEXT NOT NULL UNIQUE,
                password TEXT NOT NULL,
                session TEXT,
                created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP
            )
        ''')
        logger.info("Users table created.")

        # Create boards table if it doesn't exist
        logger.info("Creating boards table if it doesn't exist...")
        self.cursor.execute('''
            CREATE TABLE IF NOT EXISTS boards (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                name TEXT NOT NULL UNIQUE,
                owner_id INTEGER NOT NULL,
                description TEXT,
                created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                FOREIGN KEY (owner_id) REFERENCES users (id) ON DELETE SET NULL
            )
        ''')
        logger.info("Boards table created.")

        # Create posts table if it doesn't exist
        logger.info("Creating posts table if it doesn't exist...")
        self.cursor.execute('''
            CREATE TABLE IF NOT EXISTS posts (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                user_id INTEGER NOT NULL,
                board_id INTEGER NOT NULL,
                content TEXT NOT NULL,
                created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                refence INTEGER,
                FOREIGN KEY (refence) REFERENCES posts (id) ON DELETE CASCADE,
                FOREIGN KEY (user_id) REFERENCES users (id) ON DELETE SET NULL,
                FOREIGN KEY (board_id) REFERENCES boards (id) ON DELETE CASCADE
            )
        ''')
        logger.info("Posts table created.")


    def close(self):
        # Close the database connection
        logger.info("Closing database connection...")
        self.connection.commit()
        self.connection.close()
        logger.info("Database connection closed.")


    def create_user(self, username, password):
        logger.info(f"Creating user: {username}")

        # Check if the user already exists
        logger.info(f"Checking if user {username} already exists...")
        self.cursor.execute('''
            SELECT * FROM users WHERE username = ?
        ''', (username,))
        user = self.cursor.fetchone()

        if user:
            logger.warning(f"User {username} already exists.")
            return False
        logger.info(f"User {username} does not exist.")

        # Create a new user
        self.cursor.execute('''
            INSERT INTO users (username, password)
            VALUES (?, ?)
        ''', (username, password))

        self.connection.commit()
        logger.info(f"User {username} created.")
        return True
    

    def get_user(self, username):
        # Get user by username
        logger.info(f"Getting user {username}...")
        self.cursor.execute('''
            SELECT * FROM users WHERE username = ?
        ''', (username,))
        user = self.cursor.fetchone()

        if user:
            logger.info(f"User {username} found.")
            return user
        logger.warning(f"User {username} not found.")
        return None
    

    def create_board(self, name, description, owner_id):
        logger.info(f"Creating board: {name}")

        # Check if the board already exists
        logger.info(f"Checking if board {name} already exists...")
        self.cursor.execute('''
            SELECT * FROM boards WHERE name = ?
        ''', (name,))
        board = self.cursor.fetchone()

        if board:
            logger.warning(f"Board {name} already exists.")
            return False
        logger.info(f"Board {name} does not exist.")

        # Check if the owner exists
        logger.info(f"Checking if owner {owner_id} exists...")
        self.cursor.execute('''
            SELECT * FROM users WHERE id = ?
        ''', (owner_id,))
        owner = self.cursor.fetchone()

        if not owner:
            logger.warning(f"Owner {owner_id} does not exist.")
            return False
        logger.info(f"Owner {owner_id} exists.")

        # Create a new board
        self.cursor.execute('''
            INSERT INTO boards (name, owner_id, description)
            VALUES (?, ?, ?)
        ''', (name, owner_id, description))

        self.connection.commit()
        logger.info(f"Board {name} created.")
        return True
    

    def get_board(self, name):
        # Get board by name
        logger.info(f"Getting board {name}...")
        self.cursor.execute('''
            SELECT * FROM boards WHERE name = ?
        ''', (name,))
        board = self.cursor.fetchone()

        if board:
            logger.info(f"Board {name} found.")
            return board
        logger.warning(f"Board {name} not found.")
        return None
    

    def create_post(self, user_id, board_id, content, ref=None):
        logger.info(f"Creating post for user {user_id} in board {board_id}...")

        # Check if the user exists
        logger.info(f"Checking if user {user_id} exists...")
        self.cursor.execute('''
            SELECT * FROM users WHERE id = ?
        ''', (user_id,))
        user = self.cursor.fetchone()

        if not user:
            logger.warning(f"User {user_id} does not exist.")
            return False
        logger.info(f"User {user_id} exists.")

        # Check if the board exists
        logger.info(f"Checking if board {board_id} exists...")
        self.cursor.execute('''
            SELECT * FROM boards WHERE id = ?
        ''', (board_id,))
        board = self.cursor.fetchone()

        if not board:
            logger.warning(f"Board {board_id} does not exist.")
            return False
        logger.info(f"Board {board_id} exists.")

        # Check if the reference post exists
        if ref is not None:
            logger.info(f"Checking if reference post {ref} exists...")
            self.cursor.execute('''
                SELECT * FROM posts WHERE id = ?
            ''', (ref,))
            reference_post = self.cursor.fetchone()

            if not reference_post:
                logger.warning(f"Reference post {ref} does not exist.")
                return False
            logger.info(f"Reference post {ref} exists.")
        else:
            logger.info("No reference post provided.")
            
        # Create a new post
        self.cursor.execute('''
            INSERT INTO posts (user_id, board_id, content, refence)
            VALUES (?, ?, ?, ?)
        ''', (user_id, board_id, content, ref))
        self.connection.commit()
        logger.info(f"Post created for user {user_id} in board {board_id}.")
        post_id = self.cursor.lastrowid
        logger.info(f"Post ID: {post_id}")
        return post_id
    

    def get_post(self, post_id):
        # Get post by ID
        logger.info(f"Getting post {post_id}...")
        self.cursor.execute('''
            SELECT * FROM posts WHERE id = ?
        ''', (post_id,))
        post = self.cursor.fetchone()

        if post:
            logger.info(f"Post {post_id} found.")
            return post
        logger.warning(f"Post {post_id} not found.")
        return None


    def get_post_references(self, post_id, limit=10, offset=0):
        # Get references for a post
        logger.info(f"Getting references for post {post_id}...")
        self.cursor.execute('''
            SELECT * FROM posts WHERE refence = ? LIMIT ? OFFSET ?
        ''', (post_id, limit, offset))
        references = self.cursor.fetchall()

        if references:
            logger.info(f"References found for post {post_id}.")
            return references
        logger.warning(f"No references found for post {post_id}.")
        return None
    

    def get_posts(self, board_id, limit, offset):
        if board_id:
            logger.info(f"Getting posts for board {board_id}...")

            # Check if the board exists
            logger.info(f"Checking if board {board_id} exists...")
            self.cursor.execute('''
                SELECT * FROM boards WHERE id = ?
            ''', (board_id,))
            board = self.cursor.fetchone()

            if not board:
                logger.warning(f"Board {board_id} does not exist.")
                return None
            logger.info(f"Board {board_id} exists.")

            # Get posts for the board
            logger.info(f"Getting posts for board {board_id}...")
            self.cursor.execute('''
                SELECT * FROM posts WHERE board_id = ? ORDER BY created_at DESC LIMIT ? OFFSET ?
            ''', (board_id, limit, offset))
            posts = self.cursor.fetchall()

            if posts:
                logger.info(f"Posts found for board {board_id}.")
                return posts
            logger.warning(f"No posts found for board {board_id}.")
            return None
        
        else:
            # Get all posts
            logger.info(f"Getting all posts...")
            self.cursor.execute('''
                SELECT * FROM posts ORDER BY created_at DESC LIMIT ? OFFSET ?
            ''', (limit, offset))
            posts = self.cursor.fetchall()

            if posts:
                logger.info(f"Posts found.")
                return posts
            logger.warning(f"No posts found.")
            return None
    

    def get_latest_posts(self, limit):
        logger.info(f"Getting latest posts...")

        # Get latest posts
        logger.info(f"Getting latest posts...")
        self.cursor.execute('''
            SELECT * FROM posts ORDER BY created_at DESC LIMIT ?
        ''', (limit,))
        posts = self.cursor.fetchall()

        if posts:
            logger.info(f"Latest posts found.")
            return posts
        logger.warning(f"No latest posts found.")
        return None
    

    def get_boards(self, limit, offset):
        logger.info(f"Listing boards with limit {limit} and offset {offset}...")

        # Get boards with pagination
        self.cursor.execute('''
            SELECT * FROM boards ORDER BY created_at DESC LIMIT ? OFFSET ?
        ''', (limit, offset))
        boards = self.cursor.fetchall()

        if boards:
            logger.info(f"Boards found.")
            return boards
        logger.warning(f"No boards found.")
        return None
    

    def post_to_dict(self, post):
        # Convert post to dictionary
        logger.info(f"Converting post to dictionary...")
        data = {
            'id': post[0],
            'content': post[3],
            'created_at': post[4],
            'url': f"/post/{post[0]}",
        }

        # Get user information
        logger.info(f"Getting user information for post {post[0]}...")
        self.cursor.execute('''
            SELECT * FROM users WHERE id = ?
        ''', (post[1],))
        user = self.cursor.fetchone()

        if user:
            data['user'] = {
                'id': user[0],
                'name': user[1],
                'created_at': user[3]
            }
            logger.info(f"User information found for post {post[0]}.")
        else:
            logger.warning(f"User information not found for post {post[0]}.")
            data['user'] = {
                'id': None,
                'name': None,
                'created_at': None
            }
            
        # Get board information
        logger.info(f"Getting board information for post {post[0]}...")
        self.cursor.execute('''
            SELECT * FROM boards WHERE id = ?
        ''', (post[2],))
        board = self.cursor.fetchone()

        if board:
            data['board'] = {
                'id': board[0],
                'name': board[1],
                'description': board[3],
                'created_at': board[4]
            }
            logger.info(f"Board information found for post {post[0]}.")
        else:
            logger.warning(f"Board information not found for post {post[0]}.")
            data['board'] = {
                'id': None,
                'name': None,
                'description': None,
                'created_at': None
            }

        # Get reference post information
        if post[5] is not None:
            logger.info(f"Getting reference post information for post {post[0]}...")
            self.cursor.execute('''
                SELECT * FROM posts WHERE id = ?
            ''', (post[5],))
            reference_post = self.cursor.fetchone()

            if reference_post:
                data['reference'] = {
                    'id': reference_post[0],
                    'content': reference_post[3],
                    'created_at': reference_post[4]
                }
                logger.info(f"Reference post information found for post {post[0]}.")
            else:
                logger.warning(f"Reference post information not found for post {post[0]}.")
                data['reference'] = {
                    'id': None,
                    'content': None,
                    'created_at': None
                }
        else:
            logger.info("No reference post.")
            data['reference'] = None

        logger.info(f"Post converted to dictionary.")
        return data
    

    def create_session(self, username, password):
        logger.info(f"Creating session for user {username}...")

        # Check if the user exists
        logger.info(f"Checking if user {username} exists...")
        self.cursor.execute('''
            SELECT * FROM users WHERE username = ?
        ''', (username,))
        user = self.cursor.fetchone()

        if not user:
            logger.warning(f"User {username} does not exist.")
            return None
        logger.info(f"User {username} exists.")

        # Check if the password is correct
        logger.info(f"Checking password for user {username}...")
        if user[2] != password:
            logger.warning(f"Incorrect password for user {username}.")
            return None
        logger.info(f"Password for user {username} is correct.")

        # Create a new session and overwrite the old one if it exists
        session = os.urandom(16).hex()
        self.cursor.execute('''
            UPDATE users SET session = ? WHERE username = ?
        ''', (session, username))

        self.connection.commit()
        logger.info(f"Session created for user {username}.")
        return session
    

    def get_session(self, session):
        logger.info(f"Getting session {session}...")

        # Get session information
        self.cursor.execute('''
            SELECT * FROM users WHERE session = ?
        ''', (session,))
        user = self.cursor.fetchone()

        if user:
            logger.info(f"Session {session} found.")
            return user
        logger.warning(f"Session {session} not found.")
        return None
    

    def delete_session(self, session):
        logger.info(f"Deleting session {session}...")

        # Delete session
        self.cursor.execute('''
            UPDATE users SET session = NULL WHERE session = ?
        ''', (session,))

        self.connection.commit()
        logger.info(f"Session {session} deleted.")
        return True
    

    def user_to_dict(self, user):
        # Convert user to dictionary
        logger.info(f"Converting user to dictionary...")
        data = {
            'id': user[0],
            'name': user[1],
            'created_at': user[4]
        }
        logger.info(f"User converted to dictionary.")
        return data
    

    def get_user_posts(self, user_id, limit=10, offset=0):
        logger.info(f"Getting posts for user {user_id}...")

        # Get posts for the user
        self.cursor.execute('''
            SELECT * FROM posts WHERE user_id = ? ORDER BY created_at DESC LIMIT ? OFFSET ?
        ''', (user_id, limit, offset))
        posts = self.cursor.fetchall()

        if posts:
            logger.info(f"Posts found for user {user_id}.")
            return posts
        logger.warning(f"No posts found for user {user_id}.")
        return None
    

    def board_to_dict(self, board):
        # Convert board to dictionary
        logger.info(f"Converting board to dictionary...")
        data = {
            'id': board[0],
            'name': board[1],
            'description': board[3],
            'created_at': board[4]
        }
        logger.info(f"Board converted to dictionary.")
        return data
    

    def get_board_posts(self, board_id, limit=10, offset=0):
        logger.info(f"Getting posts for board {board_id}...")

        # Get posts for the board
        self.cursor.execute('''
            SELECT * FROM posts WHERE board_id = ? ORDER BY created_at DESC LIMIT ? OFFSET ?
        ''', (board_id, limit, offset))
        posts = self.cursor.fetchall()

        if posts:
            logger.info(f"Posts found for board {board_id}.")
            return posts
        logger.warning(f"No posts found for board {board_id}.")
        return None
    

    def get_all_boards_for_post(self):
        logger.info(f"Getting all boards...")

        # Get all boards
        self.cursor.execute('''
            SELECT * FROM boards
        ''')
        boards = self.cursor.fetchall()

        if boards:
            logger.info(f"Boards found.")
            # Convert boards to dictionary
            boards = [{"id": board[0], "name": board[1]} for board in boards]
            return boards
        logger.warning(f"No boards found.")
        return None